//-----------------------------------------------------------------------------
//                            CarthaGene
//
// Martin Bouchez, Patrick Chabrier, Christine Gaspin and Thomas Schiex. 
// Unite de Biometrie et Intelligence Artificielle, Toulouse, France
//
// Copyright 1997 INRA (Institut National de la Recherche Agronomique).
// This file is distributed under the terms of the Q Public License version 1.0.
// 
// $Id: BJS_RHD.h,v 1.7 2003/01/06 13:50:17 bouchez Exp $
//
// Description : Prototype de la classe BJS_RHD.
// Divers : 
//-----------------------------------------------------------------------------

#ifndef _BJS_RHD_H
#define _BJS_RHD_H

#include "BioJeuSingle.h"

#define EM_MAX_THETA_HR 0.9995
#define EM_MIN_THETA_HR 0.0001
#define EM_MAX_RETAIN   0.9999
#define EM_MIN_RETAIN   0.0001

/** Classe des jeux de donnes hybrides irradis */
class BJS_RHD : public BioJeuSingle {
 public:

  /** Constructeur. */
  BJS_RHD();

  /** Constructeur.
      @param cartag accs  l'ensemble des informations du systme. 
      @param id le numro du jeu.
      @param cross type de jeu de donne.
      @param nm nombre de marqueurs.
      @param taille de l'chantillon
      @param bitjeu champ de bit du jeu de donnes.
      @param indmarq le vecteur d'indice de marqueur
      @param echantil la matrice des informations
    */
  BJS_RHD(CartaGenePtr cartag,
	 int id,
	 charPtr nomjeu,
	 CrossType cross,
	 int nm, 
	 int taille,
	 int bitjeu,
	 int *indmarq,
	 Obs **echantil);
  
  /** Destructeur. */
  ~BJS_RHD();

  /** Compatibilite entre deux echantillons d'un marqueur
      @param numarq1 le numero du marqueur 1
      @param numarq2 le numero du marqueur 2
  */
  int Compatible(int numarq1, int numarq2) const;

  /** Fusion de deux marqueurs pour la creation d'un haplotype. Le
      marqueur 1 est utilise pour stocker le resultat.
      @param numarq1 marqueur 1 
      @param numarq2 marqueur 2 
      @return nothing.  
  */
  void Merge(int marq1,int marq2) const;

  /** Etape d'Expectation de EM.
      @param data la carte
      @param expected le vecteur d'expectation
      @return le loglike
  */
  double ComputeExpected(const Carte *data, double *expected);

 private:
  /** Calcul stat. suffisantes pour le 2pt
      @param m1 premier marqueur
      @param m2 second marqueur
      @param ss statistiques suffisantes (n[1..3])
  */
  void Prepare2pt(int m1, int m2,  int *ss) const;

  /** Calcul des estimateurs 2pt de vrais. max
      @param par vecteur des parametres du modele (theta,r mis a jour)
      @param ss statistiques suffisantes (n[1..3])
  */
  void Estimate2pt(double *par, int *ss) const;

  /** Calcul de la vraisemblance 2pt etant donnes les parametres du
      modele.
      @param par vecteur des parametres du modele (theta,r)
      @param ss statistiques suffisantes (n[1..3])
  */
  double LogLike2pt(double *par, int *ss) const;

  /** passage  priv d'une virtuelle hrite. 
  */
  double ComputeOneTwoPoints(int m1, 
			     int m2, 
			     double epsilon,
			     double *fr) const;
    
  /** passage  priv d'une virtuelle hrite. 
  */
  double LogInd(int m1, 
		int m2,
		int *nbdata) const { return 0.0;};
  
  /** passage  priv d'une virtuelle hrite.    
  */
  double EspRec(int m1,
		int m2, 
		double theta, 
		double *loglike) const { return 0.0;};

  /** Prparation des structures pour la programmation dynamique de EM.
      @param data la carte
  */
  void PreparEM(const Carte *data);

  /** nettoyage des structures pour la programmation dynamique de EM.
      @param data la carte
    */
  void NetEM(const Carte *data);

  /** Pour la programmation dynamique. tableaux des probabilits
      conditionnelles de passage par un sommet depuis la gauche
      
      Comme les proba conditionnelles sont utilises, on sait que la proba
      de passer par le sommet 1 est 1 moins la proba de passer par
      l'autre sommet. Il est donc inutile de conserver les deux probas !
      Seule celle correspondant au sommet 1 est conserve.
  */
  doublePtr SourceTo[4];

  /**  la programmation dynamique. tableaux des probabilits
       conditionnelles de passage par un sommet depuis la droite
  */
  doublePtr ToSink[4];

  /** Calcul de la probabilite de passage par une arete suivant la paire de genotypes relies
      @param a le premier gnotype
      @param b le second gnotype
      @param p la proba de cassure
      @param r la proba de rtension
      @return la probabilit
  */
  double ProbArete(int a, int b, double p, double r) const;

  /** Calcul de toutes les probas
      @param p la proba de cassure
      @param r la proba de rtension     
  */
  void CompProbArete(double p, double r);

  /**  les proba : 2,1,0 cassures & 2,1 rtension
  */
  double ProbAr2[4][4];
  double ProbAr[4][4];
  double ProbArNB[4][4];
  double ProbArRet2[4][4];
  double ProbArRet[4][4];
  
  /** Calcul des probabilits d'existence d'un chemin passant par un sommet depuis chacune des extrmits.
      @param Ind
      @param map la carte
  */
  inline double ComputeSourceTo(int Ind, const Carte *map) const;

  /** Calcul des probabilites d'existence d'un chemin passant par un sommet depuis l'extremite droite (puit)
      @param Ind
      @param map la carte
  */
  inline double ComputeToSink(int Ind, const Carte *map) const;
  
  /** Etape elementaire d'  Expectation  de EM.
      @param Ind
      @param map la carte
      @param expected le vecteur d'expectation
  */
  void ComputeOneExpected(int Ind, 
		     const Carte* map, 
		     double *expected);

  /** Calcul de la distance
      @param le theta.
      @return la distance
  */
  inline double Distance(double theta) const 
    {
      return (Theta2Ray(theta));
    };
  /** Calcul de la distance
      @param l'unit
      @param le theta.
      @return la distance
  */
  inline double Distance(char unit[2], double theta) const 
    {
      return (Theta2Ray(theta));
    }; 
};

#endif /* _BJS_RHD_H */
