//-----------------------------------------------------------------------------
//                            CarthaGene
//
// Martin Bouchez, Patrick Chabrier, Christine Gaspin and Thomas Schiex. 
// Unite de Biometrie et Intelligence Artificielle, Toulouse, France
//
// Copyright 1997 INRA (Institut National de la Recherche Agronomique).
// This file is distributed under the terms of the Q Public License version 1.0.
// 
// $Id: BioJeu.h,v 1.24 2003/02/20 13:46:37 chabrier Exp $
//
// Description : Prototype de la classe de base BioJeu.
// Divers : 
//-----------------------------------------------------------------------------

#ifndef _BIOJEU_H
#define _BIOJEU_H

#include "System.h"
#include "CGtypes.h"
#include "Genetic.h"
#include "Carte.h"

#include <stdio.h>

/** Esperance du nombre de cassures (extremites gauches) */
#define EBreak expected[NbM - 1]

/** Esperance du nombre d'extremites gauches retenues */
#define ERetained expected[NbM]

const double EM_MAX_RETAIN  =  0.9999;
const double  EM_MIN_RETAIN   = 0.0001;

class CartaGene;

typedef CartaGene *CartaGenePtr;

const double Close2Zero=1e-10;

/** Classe mre des jeux de donnes. (Tas de Bio Jeu tu sais!) */ 
class BioJeu {

 public:

  /** Les seuils de convergence - variables de classe */
  static double Epsilon1;
  static double Epsilon2;

  /** le numro du jeu */
  int Id;

  /** Type de population. */
  CrossType Cross;

  /** Champ de bit. */
  int BitJeu;
  
  /** Nombre de marqueurs. */
  int NbMarqueur;

  /** Nombre d'individus. */
  int TailleEchant;

  /** Nombre de Mioses. */
  int NbMeiose;
  
  /** Compteur du nombre d'appels aux fonctions EM.
      Il est li au remplissage d'un Tas.
    */
  unsigned int NbEMCall;
  unsigned int NbEMHit;
  unsigned int NbEMUnconv;

  /** 
   */
  double Em_Max_Theta;
  double Em_Min_Theta;
  /** Constructeur. */
  BioJeu();

  /** Constructeur.
      @param cartag accs  l'ensemble des informations du systme. 
      @param id le numro du jeu.
      @param cross type de jeu de donne.
      @param nm nombre de marqueurs.
      @param bitjeu champ de bit du jeu de donnes.
  */
  BioJeu(CartaGenePtr cartag,
	 int id,
	 CrossType cross,
	 int nm, 
	 int bitjeu);

  /** Destructeur. */
  virtual ~BioJeu();

  /** Retourne un champ des jeux fusionns sur l'ordre(!=CON).
      @return un champ de bits
  */
  virtual int GetBJMO() const = 0;


  /** Calcul du log sous l'hypothse d'indpendance 2pt.
      @param m1 le numro du premier marqueur
      @param m2 le numro du second marqueur
      @param nbdata le nombre de donnes informatives(o)
      @return le log
  */
  virtual double LogInd(int m1, 
			int m2,
			int *nbdata) const = 0;

  /** Calcul l'esprance du nombre de recombinants, (log volutif).
      @param m1 le numro du premier marqueur
      @param m2 le numro du second marqueur
      @param theta
      @param loglike(o)
      @return l'esprance      
  */
  virtual double EspRec(int m1,
			int m2, 
			double theta, 
			double *loglike) const = 0;


  /** Test de compatibilite de deux marqueurs pour la creation d'un haplotype
      @param numarq1 marqueur 1
      @param numarq2 marqueur 2
      @return critre vrifi ou pas.
  */
  virtual int Compatible(int marq1,int marq2) const = 0;
  
   /** Fusion de deux marqueurs pour la creation d'un haplotype. Le
      marqueur 1 est utilise pour stocker le resultat.
      @param numarq1 marqueur 1 
      @param numarq2 marqueur 2 
      @return nothing.  
   */
  virtual void Merge(int marq1,int marq2) const = 0;
  
  /** Affichage de l'chantillon pour un marqueur
      @param numarq le numro du marqueur
  */
  virtual void DumpEchMarq(int numarq) const = 0;
  
  /** Affichage de l'chantillon complet. */
  virtual void DumpEch(void) const = 0;

  /** Affichage de la matrice des LOD scores 2pt. */
  virtual void DumpTwoPointsLOD() const = 0;

  /** Affichage de la matrice des FR 2pt. */
  virtual void DumpTwoPointsFR() const = 0;

  /** Accs  une valeur LOD score 2pt.
      @param m1 le numro du premier marqueur
      @param m2 le numro du second marqueur
      @return le lod
  */
  virtual double GetTwoPointsLOD(int m1, int m2) const = 0;

  /** Accs  la distance 2pt.
      @param m1 le numro du premier marqueur
      @param m2 le numro du second marqueur
      @return la distance
  */
  virtual double GetTwoPointsDH(int m1, int m2) const = 0;
  
  /** Accs  la fraction de recombinaison 2pt.
      @param m1 le numro du premier marqueur
      @param m2 le numro du second marqueur
      @return la fraction de recombinaison
  */
  virtual double GetTwoPointsFR(int m1, int m2) const = 0;
  
  /** Calcul du maximum de vraisemblance par EM.     
      @param data la carte
      @param epsilon le seuil de convergence
      @return le maximum de vraissemblance
  */
  virtual double ComputeEM(Carte *data);  

  /** Calcul de la vraisemblance maximum par EM.
      Une interface plus sophistique du calcul de la vraisemblance par EM:
      On converge jusqu' epsilon1. Si la vraisemblance est infrieure 
      threshold on laisse tomber, sinon on converge  epsilon2.
      @param data la carte
      @param threshold niveau a partir duquel on convergera a epsilon2
      @param epsilon1 le premier seuil de convergence
      @param epsilon2 le second seuil de convergence
      @return le maximum de vraissemblance
  */
  virtual double ComputeEMS (Carte *data, 
			     double threshold);

  /** Prparation des structures pour la programmation dynamique de EM.
      @param data la carte
  */
  virtual void PreparEM(const Carte *data) = 0;

  /** nettoyage des structures pour la programmation dynamique de EM.
      @param data la carte
  */
  virtual void NetEM(const Carte *data) = 0;  

  /** Etape d'Expectation de EM.
      @param data la carte
      @param expected le vecteur d'expectation
      @return le loglike
  */
  virtual double ComputeExpected(const Carte *data, double *expected) = 0;

  /** Etape de Maximisation de EM.
      @param data la carte
      @param expected le vecteur d'expectation
      @return le fait qu'une limite de domaine d'optimisation a ete atteinte
  */
  virtual int Maximize(Carte *data, double *expected) const;

  /** Calcul, affichage et stockage des groupes.
      On tablit un graphe dont les noeuds sont les marqueurs. Il existe une arrte entre deux marqueurs, si le critere de lod > seuil et de distance < seuil est vrifi. Puis on extrait les composantes connexes qui sont quivalentes aux groupes.
      @param disthres le seuil de distance 2pt.
      @param lodthres le seuil de LOD 2pt.
      @return le nombre de groupes.
  */
  virtual int Groupe(double disthres, double lodthres) const;

  /** Affichage simple d'une carte
      @param data la carte
  */
  virtual void PrintMap(Carte *data) const;

  /** Matrice 2p des distances
      @param unit l'unit de distance: 
      'k' = kosambi ou centiray, 
      'h' = haldane ou centiray.
  */
  virtual void PrintTwoPointsDist(char unit[2]) const;

  /** retourne une liste contenant une carte
      @param unit l'unit de distance: 
      'k' = kosambi ou centiray, 
      'h' = haldane ou centiray.
      @param data la carte
      @return une liste
  */
  virtual char *** GetMap(char unit[2], Carte *data) const;

  /** Affichage dtaill d'une carte
      @param data la carte
      @param envers  l'envers ou  l'endroit
      @param data la carte de rfrence
  */
  virtual void PrintDMap(Carte *data, int envers, Carte *dataref);

  /** Test l'existance d'un couple
      @param m1 marqueur 1
      @param m2 marqueur 2
      @return 0 ou 1
  */
  virtual int Couplex(int m1, int m2) const;

  /** Mise  0
    */
  virtual void ResetNbEMCall() {NbEMCall = NbEMHit = NbEMUnconv = 0;};
  
  /** Affichage     
    */
  virtual void PrintEMCall() {
    if (NbEMCall) {
      sprintf(bouf,"           Set %2d : %d (%d,%d)\n", Id, NbEMCall,NbEMHit,NbEMUnconv);
      pout(bouf);
    }
  };

  /** Test de l'existance d'un RH dans l'arborence
      @return 0 ou 1
  */
  virtual int HasRH() const = 0;

  /** Test de l'existance d'un RISib ou RISelf dans l'arborence
      @return 0 ou 1
  */
  virtual int HasRI() const = 0;

  /** Calcul stat. suffisantes pour le 2pt
      @param m1 premier marqueur
      @param m2 second marqueur
      @param ss statistiques suffisantes (n[1..3])
  */
  virtual void Prepare2pt(int m1, int m2,  int *ss) const = 0;

  /** Calcul des estimateurs 2pt de vrais. max
      @param par vecteur des parametres du modele (theta,r mis a jour)
      @param ss statistiques suffisantes (n[1..3])
  */
  virtual void Estimate2pt(double *par, int *ss) const =0;

  /** Calcul de la vraisemblance 2pt etant donnes les parametres du
      modele.
      @param par vecteur des parametres du modele (theta,r)
      @param ss statistiques suffisantes (n[1..3])
  */
  virtual double LogLike2pt(double *par, int *ss) const =0;

 protected:

  /** Accs  l'ensemble des informations du systme. */
  CartaGene *Cartage;
  
  /** indice local de chaque marqueur. */
  int *IndMarq;
  
  /** Taux de retention general (RH Only) */
  double OverallR;;

  /** Fraction de recombinaison 2pt. */
  doublePtr *TwoPointsFR;
  
  /** Distance selon Haldane/Ray 2pt. */
  doublePtr *TwoPointsDH;

  /** LOD score 2pt. */
  doublePtr *TwoPointsLOD;

  /** Calcul du LOD score et de fraction de recombinants 2pt.
      @param m1 le numro du premier marqueur
      @param m2 le numro du second marqueur
      @param epsilon le seuil de convergence
      @param fr la fraction de recombinants(i,o)
      @return le LOD
  */
  virtual double ComputeOneTwoPoints(int m1, 
				     int m2, 
				     double epsilon,
				     double *fr) const = 0;

  /** Calcul des valeurs de points et remplissage des matrices.
      En l'occurence, il s'agit uniquement des LOD 2pts des couples de marqueurs prsents dans les deux jeux fusionns.
  */
  virtual void ComputeTwoPoints(void) = 0;
  
  /** Parcours en profondeur(dtection des groupes)
      @param ladj liste d'adjacence
      @param visited vecteur tmoin de visite
      @param vertex le sommet
      @param la couleur
      @param plink pointeur sur la liste des groupes
  */
  void DFSVisit(NodintPtr *ladj, 
		intPtr visited, 
		int vertex, 
		int color,
		NodptrPtr plink) const;

  /** Critre de linkage entre 2 marqueurs.
      @param disthres PAS UTILISE.
      @param lodthres le seuil de LOD 2pt.
      @param k le numro du premier marqueur
      @param l le numro du second marqueur
      @return critre vrifi ou pas.
  */
  inline virtual int GroupTest(double disthres, double lodthres, int k, int l) const 
    {
      return ( GetTwoPointsDH(k, l) <= disthres && 
	       GetTwoPointsLOD(k, l) >= lodthres);
    };

  /** affichage du seuil de LOD
      @param disthres PAS UTILISE.
      @param lodthres le seuil de LOD 2pt.
  */
  inline virtual void GroupMess(double disthres, double lodthres) const
    {
      sprintf(bouf,"\nLinkage Groups :\n");
      pout(bouf);
      sprintf(bouf,"---------------:\n");
      pout(bouf);
      sprintf(bouf,"LOD threshold=%.2f\nDistance threshold=%.2f:\n", lodthres, 100.0*disthres);
      pout(bouf);
    };

  /** Calcul de la distance
      @param le theta.
      @return la distance
  */
  inline virtual double Distance(double theta) const 
    {
      return (Haldane(theta));
    };
  
  /** Calcul de la distance
      @param l'unit
      @param le theta.
      @return la distance
  */
  inline virtual double Distance(char unit[2], double theta) const 
    {
      switch(unit[0]) 
	{
	case 'h': return (Haldane(theta));
	case 'k': return (Kosambi(theta));
	}
      return (Haldane(theta));
    };
};

typedef BioJeu *BioJeuPtr;
 
#endif /* _BIOJEU_H */
